
/***************************************************************************
 *
 * IPAT - Interferometer Threshold Curve Plotter
 *
 * Stephen R. Whiteley (stevew@srware.com)
 * 
 * Copyright (c) 2002 Whiteley Research Inc.  All Rights Reserved. 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************
 * $Id: screen.cc,v 2.8 2009/01/11 21:07:41 stevew Exp $
 ***************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef WIN32_NATIVE
#include <windows.h>
#include <windowsx.h>
#else
#include <gtk/gtk.h>
#endif
#include "ipat.h"

#define sn(z) (((z) >= 0) ? 1 : 0)

extern values vals;
extern panel app;

static char *ecvt12(double);

clr Colors[] = {
    { 0, 0, 0, 0 },
    { 0, 255, 255, 255 },
    { 0, 255, 0,   0 },
    { 0, 80,  255, 80 },
    { 0, 140, 100, 255 },
    { 0, 255, 255, 128 },
    { 0, 80,  255, 255 },
    { 0, 255, 128, 255 },
    { 0, 200, 200, 200 },
    { 0, 192, 192, 64 },
    { 0, 0,   200, 0, },
    { 0, 0,   0,   180 },
    { 0, 200, 120, 80 },
    { 0, 100, 140, 160 },
    { 0, 200, 0,   200 },
    { 0, 255, 192, 64 }
};

//
// struct screen functions
//

screen::screen()
{
    X = Y = 0;
    W = H = 0;
    W2 = 0;
    H2 = H3 = 0;
    XL = XU = 0;
    YL = YU = 0;
    XS = YS = 0;
    XC = YC = 0;

    da = 0;
    gc = 0;
#ifdef WIN32_NATIVE
    curforeg = 0;
    curstyle = 0;
    font = 0;
#else
    xorgc = 0;
    font = 0;
    cmap = 0;
    bpm = 0;
    oldwin = 0;
    bpm_width = bpm_height = 0;
#endif
    last_x = last_y = 0;
    draw_ghost = false;
    first_ghost = false;
}


void
screen::clear()
{
#ifdef WIN32_NATIVE
    RECT r;
    GetClientRect(da, &r);
    r.right++;
    r.bottom++;
    HBRUSH brush = (HBRUSH)GetClassLong(da, GCL_HBRBACKGROUND);
    if (!brush) {
        brush = CreateSolidBrush(GetSysColor(COLOR_3DFACE));
        FillRect(gc, &r, brush);
        DeleteBrush(brush);
    }
    else
        FillRect(gc, &r, brush);
#else
#if GTK_CHECK_VERSION(2,0,0)
    if (!GDK_IS_PIXMAP(da->window))
#else
    if (gdk_window_get_type(da->window) != GDK_WINDOW_PIXMAP)
#endif
        gdk_window_clear(da->window);
    else {
        set_color(0);
        gdk_draw_rectangle(da->window, gc, true, 0, 0, bpm_width, bpm_height);
    }
#endif
}


void
screen::define_color(int indx, int red, int green, int blue)
{
    if (indx < 0 || indx >= 16)
        return;
#ifdef WIN32_NATIVE
    Colors[indx].pixel = RGB(red, green, blue);
#else
    GdkColor newcolor;
    newcolor.red   = (red   * 256);  
    newcolor.green = (green * 256);
    newcolor.blue  = (blue  * 256);
    newcolor.pixel = Colors[indx].pixel;
    if (gdk_colormap_alloc_color(cmap, &newcolor, false, true)) {
        Colors[indx].pixel = newcolor.pixel;
        gdk_gc_set_foreground(gc, &newcolor);
    }
#endif
}


void
screen::set_color(int indx)
{
    if (indx < 0 || indx >= 16)
        return;
#ifdef WIN32_NATIVE
    curforeg = indx;
    SetTextColor(gc, Colors[indx].pixel);
    HPEN pen = CreatePen(curstyle ? PS_DOT : PS_SOLID, 0, Colors[indx].pixel);
    pen = SelectPen(gc, pen);
    if (pen)
        DeletePen(pen);

#else
    GdkColor clr;
    clr.pixel = Colors[indx].pixel;
    gdk_gc_set_foreground(gc, &clr);
#endif
}


void
screen::pixel(int x, int y)
{
#ifdef WIN32_NATIVE
    SetPixelV(gc, x, YS - y, Colors[curforeg].pixel);
#else
    gdk_draw_point(da->window, gc, x, YS - y);
#endif
}


void
screen::line(int x1, int y1, int x2, int y2)
{
#ifdef WIN32_NATIVE
    MoveToEx(gc, x1, YS - y1, 0);
    LineTo(gc, x2, YS - y2);
#else
    gdk_draw_line(da->window, gc, x1, YS - y1, x2, YS - y2);
#endif
}


void
screen::box(int x1, int y1, int x2, int y2)
{
    y1 = YS - y1;
    y2 = YS - y2;
    if (x1 > x2) {
        int temp = x1;
        x1 = x2;
        x2 = temp;
    }
    if (y1 > y2) {
        int temp = y1;
        y1 = y2;
        y2 = temp;
    }
#ifdef WIN32_NATIVE
    RECT r;
    r.left = x1;
    r.top = y1;
    r.right = x2;
    r.bottom = y2;
    // faster?  Ousterhout claims so in tk
    COLORREF oldc = SetBkColor(gc, Colors[curforeg].pixel);
    SetBkMode(gc, OPAQUE);
    ExtTextOut(gc, 0, 0, ETO_OPAQUE, &r, 0, 0, 0);
    SetBkColor(gc, oldc);
    SetBkMode(gc, TRANSPARENT);
#else
    gdk_draw_rectangle(da->window, gc, true, x1, y1, x2 - x1 + 1,
        y2 - y1 + 1);
#endif
}


void
screen::text(const char *str, int x, int y, int)
{
#ifdef WIN32_NATIVE
    TextOut(gc, x, YS - y + 2, str, strlen(str));
#else
    gdk_draw_string(da->window, font, gc, x, YS - (y - font->descent), str);
#endif
}


void
screen::set_linestyle(int style)
{
#ifdef WIN32_NATIVE
    HPEN pen = CreatePen(style ? PS_DOT : PS_SOLID, 0,
        Colors[curforeg].pixel);
    pen = SelectPen(gc, pen);
    if (pen)
        DeletePen(pen);
    curstyle = style;
#else
    if (style == 0)
        gdk_gc_set_line_attributes(gc, 0, GDK_LINE_SOLID,
            GDK_CAP_BUTT, GDK_JOIN_MITER);
    else {
#ifdef OLDGTK
        // For gtk 1.2.6 and earlier, use this
        gchar dash[4];
#else
        gint8 dash[4];
#endif
        dash[0] = 4;
        dash[1] = 4;
        gdk_gc_set_line_attributes(gc, 0, GDK_LINE_ON_OFF_DASH,
             GDK_CAP_BUTT, GDK_JOIN_MITER);
        gdk_gc_set_dashes(gc, 0, dash, 2);
    }
#endif
}


void
screen::swap_pm(bool topm)
{
    if (!da)
        return;
#ifdef WIN32_NATIVE
#else
    if (topm) {
        int wid, hei;
        gdk_window_get_size(da->window, &wid, &hei);
        if (!bpm || wid != bpm_width || hei != bpm_height) {
            if (bpm)
                gdk_pixmap_unref(bpm);
            bpm = gdk_pixmap_new(da->window, wid, hei,
                gdk_window_get_visual(da->window)->depth);
            if (bpm) {
                bpm_width = wid;
                bpm_height = hei;
            }
            else
                return;
        }
        oldwin = da->window;
        da->window = bpm;
    }
    else {
        if (oldwin) {
            gdk_window_copy_area(oldwin, gc, 0, 0, da->window,
                0, 0, bpm_width, bpm_height);
            da->window = oldwin;
            oldwin = 0;
        }
    }
#endif
}


void
screen::update(int x, int y, int w, int h)
{
#ifdef WIN32_NATIVE
#else
    if (bpm && !oldwin)
        gdk_window_copy_area(da->window, gc, x, y, bpm, x, y, w, h);
#endif
}


// Main driver for plot generation
//
void
screen::displ()
{
    bool tmpg = draw_ghost;
    if (draw_ghost && !first_ghost)
        marker(last_x, last_y);
    first_ghost = true;
    draw_ghost = false;
    swap_pm(true);
#ifdef WIN32_NATIVE
    if (app.flags.threej) {
        EnableWindow(app.entries[2].entry, true);
        EnableWindow(app.entries[3].entry, true);
        EnableWindow(app.entries[4].entry, true);
        EnableWindow(app.entries[7].entry, true);
        EnableWindow(app.entries[2].ud, true);
        EnableWindow(app.entries[3].ud, true);
        EnableWindow(app.entries[4].ud, true);
        EnableWindow(app.entries[7].ud, true);
    }
    else {
        EnableWindow(app.entries[2].entry, false);
        EnableWindow(app.entries[3].entry, false);
        EnableWindow(app.entries[4].entry, false);
        EnableWindow(app.entries[7].entry, false);
        EnableWindow(app.entries[2].ud, false);
        EnableWindow(app.entries[3].ud, false);
        EnableWindow(app.entries[4].ud, false);
        EnableWindow(app.entries[7].ud, false);
    }
#else
    if (app.flags.threej) {
        gtk_widget_set_sensitive(app.entry_l3, true);
        gtk_widget_set_sensitive(app.entry_l4, true);
        gtk_widget_set_sensitive(app.entry_frac, true);
        gtk_widget_set_sensitive(app.entry_i3, true);
    }
    else {
        gtk_widget_set_sensitive(app.entry_l3, false);
        gtk_widget_set_sensitive(app.entry_l4, false);
        gtk_widget_set_sensitive(app.entry_frac, false);
        gtk_widget_set_sensitive(app.entry_i3, false);
    }
#endif

    if (app.flags.threej) {
        app.Period = 2*2.067930/(vals.la+vals.lb+vals.lc+vals.ld);
        scale.ymax = app.Ic = vals.ix + vals.i2 + vals.iy;
    }
    else {
        app.Period = 2.067930/(vals.la+vals.lb);
        scale.ymax = app.Ic = vals.ix + vals.i2;
    }

    if (app.flags.y) {
        scale.ymin = -scale.ymax;
        scale.yy = Y/2;
        scale.yo = YC;
    }
    else {
        scale.ymin = 0;
        scale.yy = Y;
        scale.yo = YL;
    }

    if (app.flags.x) {
        scale.xmax = app.Period*4/app.Scale;
        scale.xmin = -scale.xmax;
        scale.xmx = X/2;
        scale.xmn = scale.xmx;
        scale.xo = XC;
    }
    else {
        scale.xmax = app.Period*8/app.Scale;
        scale.xmin = 0;
        scale.xmx = X;
        scale.xmn = 0;
        scale.xo = XL;
    }
    scale.xx = app.Scale * X;

    // Draw axes
    clear();
    set_color(8);
    line(XL,YL,XU,YL);
    line(XU,YL,XU,YU);
    line(XU,YU,XL,YU);
    line(XL,YU,XL,YL);
    if (!app.flags.s) {
        set_linestyle(1);
        line(XC,YL,XC,YU);
        line(XL,YC,XU,YC);
        set_linestyle(0);
    }

    app.flags.c = 0;

    if (app.flags.z) {
        set_color(1);
        text("Pi",XL-2*W-W2,YU-H,0);
        text("-Pi",XL-3*W-W2,YL,0);
        text("-Pi",XL,YL-H-1,0);
        text("Pi",XU-2*W,YL-H-1,0);
        set_color(2);
        text("phi1",XL+6*W,YL-H-1,0);
        set_color(3);
        text("phi2",XL+12*W,YL-H-1,0);
        if (app.flags.threej) {
            set_color(4);
            text("phi3",XL+18*W,YL-H-1,0);
        }
        set_color(1);
        text("arctan(Ic/Flux)",XC-7*W,YL-H-1,0);
    }

    if (!app.flags.threej) {
        // plot the contours for 2 junctions
        twojj();
    }
    else {
        app.Order1 = app.Order2 = 0;
        threejj();
        app.Order1 = 2*PI;
        if (!app.flags.n) {
            // order 1 contours
            threejj();
        }
        app.Order2 = -2*PI;
        if (!app.flags.r) {
            // order 2 contours
            threejj();
        }
    }

    //
    // Show parameters
    //

    int y = YU + H2;
    if (!app.flags.s) {
        // Draw dots - part of axes
        set_color(1);
        int dx = X/8, dy = Y/8;
        for (int i=1; i<4; i++) {
            for (int j=1; j<4; j++) {
                pixel(XL+i*dx,YL+j*dy);
                pixel(XC+i*dx,YL+j*dy);
                pixel(XL+i*dx,YC+j*dy);
                pixel(XC+i*dx,YC+j*dy);
            }
        }
    }

    char buf[128];
    set_color(8);
    if (!app.flags.z) {
        sprintf(buf,"Period %g mA, Coupling %g",app.Period/app.Coupling,
            app.Coupling);
        text(buf,XL,YL - H3,0);
    }
    if (app.flags.threej) {
        sprintf(buf,"L1  = %g", vals.la);
        text(buf,XL,y+H,0);
        sprintf(buf,"L2  = %g", vals.lb);
        text(buf,XL,y,0);
        sprintf(buf,"L3  = %g", vals.lc);
        text(buf,XL+X/4,y+H,0);
        sprintf(buf,"L4  = %g", vals.ld);
        text(buf,XL+X/4,y,0);
        sprintf(buf,"Ic1 = %g", vals.ix);
        text(buf,XC,y+H,0);
        sprintf(buf,"Ic2 = %g", vals.i2);
        text(buf,XC,y,0);
        sprintf(buf,"Ic3 = %g", vals.iy);
        text(buf,XC+X/4,y+H,0);
        sprintf(buf,"Ileft/Ig = %g", vals.lfrac);
        text(buf,XC+X/4,y,0);
        set_color(1);
        if (app.flags.z)
            text("3 JUNCTION INTERFEROMETER - PHASE",XL,YL-H3-H,0);
        else
            text("3 JUNCTION INTERFEROMETER - THRESHOLD",XL,YL-H3-H,0);
    }
    else {
        sprintf(buf,"L1  = %g", vals.la);
        text(buf,XL,y,0);
        sprintf(buf,"L2  = %g", vals.lb);
        text(buf,XL+X/4,y,0);
        sprintf(buf,"Ic1 = %g", vals.ix);
        text(buf,XC,y,0);
        sprintf(buf,"Ic2 = %g", vals.i2);
        text(buf,XC+X/4,y,0);
        set_color(1);
        if (app.flags.z)
            text("2 JUNCTION INTERFEROMETER - PHASE",XC-16*W,y+H,0);
        else
            text("2 JUNCTION INTERFEROMETER - THRESHOLD",XC-18*W,y+H,0);
    }
    plpts();
    refmark();
    swap_pm(false);
    draw_ghost = tmpg;
}


// Display curve for two junction interferometer
//
void
screen::twojj()
{
    /*********************************************************************

        Basic equations:

        (1)  It = Ic1*sin(phi1) + Ic2*sin(phi2)

        (2)  FLX = phi1 - phi2 + PHI*( L1*Ic1*sin(phi1)  -
                 L2*Ic2*sin(phi2) )

        (3)  1/(Ic1*cos(phi1)) + 1/(Ic2*cos(phi2) + PHI*(L1+L2) = 0

        We loop over phi1, then phi2 (to better fill in curve), using
        (3) to get the other angle. (1) and (2) give the coordinate to plot.

    **********************************************************************/

    double f1 = PHI*(vals.la + vals.lb);
    for (double phi2 = -PI; phi2 <= PI; phi2 += .01) {

        double f2 = vals.i2*cos(phi2);
        if (f2 == 0) continue;
        f2 = 1/f2;
        double f3 = vals.ix*(f1 + f2);
        if (f3 == 0) continue;
        f3 = -1/f3;
        if (f3 > 1 || f3 < -1) continue;
        if (fabs(phi2) > .5*PI && f3 < 0) continue;
        double phi1 = acos(f3);
        double s1 = vals.ix*sin(phi1);
        double s2 = vals.i2*sin(phi2);
        double ic = s1 + s2;
        double flx = phi1 - phi2 + PHI*(vals.la*s1 - vals.lb*s2);
        if (app.flags.z)
            plotphi(phi1,phi1,phi2,flx,ic);
        else
            plot(flx/(16*PI), ic/app.Ic);
    }
    for (double phi1 = -PI; phi1 <= PI; phi1 += .01) {

        double f2 = vals.ix*cos(phi1);
        if (f2 == 0) continue;
        f2 = 1/f2;
        double f3 = vals.i2*(f1 + f2);
        if (f3 == 0) continue;
        f3 = -1/f3;
        if (f3 > 1 || f3 < -1) continue;
        if (fabs(phi1) > .5*PI && f3 < 0) continue;
        double phi2 = acos(f3);
        double s1 = vals.ix*sin(phi1);
        double s2 = vals.i2*sin(phi2);
        double ic = s1 + s2;
        double flx = phi1 - phi2 + PHI*(vals.la*s1 - vals.lb*s2);
        if (app.flags.z)
            plotphi(phi1,phi1,phi2,flx,ic);
        else
            plot(flx/(16*PI), ic/app.Ic);
    }
}


// display curves for three junction interferometer
//
void
screen::threejj()
{
    /*********************************************************************

  Basic equations:
  sN = IcN*sin(phiN)
  cN = icN*cos(phiN)

  (1)  It = s1 + s2 + s3
          = Il + Ir, Il/It = C  ( C is splitting fraction )
       (gate current)

  (2)  f1 = phi1 - phi2 + PHI*(L1*s1 - L2*(C(s2 + s3) + (C-1)*s1)) + Order1
       (left loop flux)

       f2 = phi2 - phi3 + PHI*(L3*((1-C)*(s1 + s2) - C*s3) - l4*s3) + Order2
       (right loop flux)

  (3)  I~ = It + A*(f1 terms) + B*(f2 terms)
       (lagrange multipliers)

  (4)  Partial derivatives

       phi1:
       c1 +
          A * (1 + PHI*L1*c1 + (1-C)*PHI*L2*c1) +
          B * (1-C)*PHI*L3*c1 = 0

       phi2:
       c2 -
          A * (1 + C*PHI*L2*c2) +
          B * (1 + (1-C)*PHI*L3*c2) = 0

       phi3:
       c3 -
          A * C*PHI*L2*c3 -
          B * (1 + PHI*L4*c3 + C*PHI*L3*c3) = 0

  (5)  FLX = f1 + f2
       (total external flux)

  (6)  (L3+L4)*f1 - (L1+L2)*f2 = 0
       (assume the flux is coupled through the loop inductance)
       A common control line divides the flux between the two loops as
       above.

    **********************************************************************/

#define PHISTEP .0025*PI

    double start = 0.0;
    for (;;) {
        int count = 0;
        bool found1 = true;
        bool found2 = true;
        double phi1 = start;
        double phi2 = 0.0,  phi3 = 0.0;
        for (;;) {

            if (found1)
                found1 = solve13(phi1,&phi2);
            if (found2)
                found2 = solve12(phi1,&phi3);
            if (!found1 && !found2)
                break;
            phi1 += PHISTEP;
            if (phi1 > PI)
                break;
            count++;
        }

        found1 = true;
        found2 = true;
        phi1 = start;
        phi2 = phi3 = 0;
        for (;;) {

            if (found1)
                found1 = solve13(phi1,&phi2);
            if (found2)
                found2 = solve12(phi1,&phi3);
            if (!found1 && !found2)
                break;
            phi1 -= PHISTEP;
            if (phi1 < -PI)
                break;
            count++;
        }
        if (count > 10) break;
        if (start == 0) {
            start = PI/2;
            continue;
        }
        if (start == PI/2) {
            start = -start;
            continue;
        }
        break;
    }

    start = 0;
    for (;;) {
        int count = 0;
        bool found1 = true;
        bool found2 = true;
        double phi3 = start;
        double phi1 = 0.0, phi2 = 0.0;
        for (;;) {

            if (found1)
                found1 = solve31(phi3,&phi2);
            if (found2)
                found2 = solve32(phi3,&phi1);
            if (!found1 && !found2)
                break;
            phi3 += PHISTEP;
            if (phi3 > PI)
                break;
            count++;
        }

        found1 = true;
        found2 = true;
        phi3 = start;
        phi1 = phi2 = 0;
        for (;;) {

            if (found1)
                found1 = solve31(phi3,&phi2);
            if (found2)
                found2 = solve32(phi3,&phi1);
            if (!found1 && !found2)
                break;
            phi3 -= PHISTEP;
            if (phi3 < -PI)
                break;
            count++;
        }
        if (count > 10) break;
        if (start == 0) {
            start = PI/2;
            continue;
        }
        if (start == PI/2) {
            start = -start;
            continue;
        }
        break;
    }

    start = 0;
    for (;;) {
        int count = 0;
        bool found1 = true;
        bool found2 = true;
        double phi2 = start;
        double phi1 = 0.0, phi3 = 0.0;
        for (;;) {

            if (found1)
                found1 = solve21(phi2,&phi3);
            if (found2)
                found2 = solve23(phi2,&phi1);
            if (!found1 && !found2)
                break;
            phi2 += PHISTEP;
            if (phi2 > PI)
                break;
            count++;
        }

        found1 = true;
        found2 = true;
        phi2 = start;
        phi1 = phi3 = 0;
        for (;;) {

            if (found1)
                found1 = solve21(phi2,&phi3);
            if (found2)
                found2 = solve23(phi2,&phi1);
            if (!found1 && !found2)
                break;
            phi2 -= PHISTEP;
            if (phi2 < -PI)
                break;
            count++;
        }
        if (count > 10) break;
        if (start == 0) {
            start = PI/2;
            continue;
        }
        if (start == PI/2) {
            start = -start;
            continue;
        }
        break;
    }
}


#define LIMIT(p) \
    if (p > PI) p -= 2*PI; \
    else if (p < -PI) p += 2*PI

bool
screen::solve12(double phi1, double *phi3p)
{
    double phi3 = *phi3p - PI*1e-4;
    double lastphi = phi3;
    double s1 = vals.ix*sin(phi1);

    double lastz = 0.0;
    for (int count = 0; count < NITER; count++) {

        double phi2;
        if (vals.get2(phi1,phi3,&phi2)) break;
    
        double s2 = vals.i2*sin(phi2);
        double s3 = vals.iy*sin(phi3);

        double f1 = phi1 + PHI*(vals.la*s1 -
            vals.lb*((vals.lfrac-1)*s1 + vals.lfrac*s3)) + app.Order1;

        double f2 = -phi3 + PHI*(vals.lc*(-vals.lfrac*s3 + (1-vals.lfrac)*s1)
            - vals.ld*s3) + app.Order2;

        double f3 = -phi2 - vals.lfrac*PHI*vals.lb*s2;

        double f4 = phi2 + (1-vals.lfrac)*PHI*vals.lc*s2;

        double z = (vals.lc + vals.ld)*f1 - (vals.la + vals.lb)*f2;
        double zz = (vals.lc + vals.ld)*f3 - (vals.la + vals.lb)*f4;

        if (sn(zz) != sn(z)) {
            f1 += f3;
            f2 += f4;
            z += zz;
        }
        else {
            f1 -= f3;
            f2 -= f4;
            z -= zz;
            phi2 = -phi2;
            s2 = -s2;
        }

        if (count > 0) {
            if (z != lastz) {
                double phinew = phi3 - (phi3 - lastphi)*z/(z-lastz);
                LIMIT(phinew);
                lastphi = phi3;
                phi3 = phinew;
            }
            else
                count = NITER;

            if (fabs(z) < .001) {
                if (!vals.solcheck(phi1,phi3)) break;
                double I = s1 + s2 + s3;
                double FLX = f1 + f2;
                if (app.flags.z)
                    plotphi(phi1,phi2,phi3,FLX,I);
                else
                    plot(FLX/(32*PI), I/app.Ic);
                *phi3p = phi3;
                return (true);
            }
        }
        else {
            phi3 = *phi3p;
        }
        lastz = z;
    }
    return (false);
}


bool
screen::solve13(double phi1, double *phi2p)
{
    double phi2 = *phi2p - PI*1e-4;
    double lastphi = phi2;
    double s1 = vals.ix*sin(phi1);

    double lastz = 0.0;
    for (int count = 0; count <= NITER; count++) {

        double phi3;
        if (vals.get3(phi1,phi2,&phi3)) break;

        double s2 = vals.i2*sin(phi2);
        double s3 = vals.iy*sin(phi3);

        double f1 = phi1 - phi2 + PHI*(vals.la*s1 -
            vals.lb*((vals.lfrac-1)*s1 + vals.lfrac*s2)) + app.Order1;

        double f2 = phi2 + PHI*((1-vals.lfrac)*vals.lc*(s2 + s1)) + app.Order2;

        double f3 = -vals.lfrac*PHI*vals.lb*s3;

        double f4 = -phi3 -  PHI*(vals.ld + vals.lfrac*vals.lc)*s3;

        double z = (vals.lc + vals.ld)*f1 - (vals.la + vals.lb)*f2;
        double zz = (vals.lc + vals.ld)*f3 - (vals.la + vals.lb)*f4;

        if (sn(zz) != sn(z)) {
            f1 += f3;
            f2 += f4;
            z += zz;
        }
        else {
            f1 -= f3;
            f2 -= f4;
            z -= zz;
            phi3 = -phi3;
            s3 = -s3;
        }

        if (count > 0) {
            if (z != lastz) {
                double phinew = phi2 - (phi2 - lastphi)*z/(z-lastz);
                LIMIT(phinew);
                lastphi = phi2;
                phi2 = phinew;
            }
            else
                count = NITER;

            if (fabs(z) < .001) {
                if (!vals.solcheck(phi1,phi3)) break;
                double I = s1 + s2 + s3;
                double FLX = f1 + f2;
                if (app.flags.z)
                    plotphi(phi1,phi2,phi3,FLX,I);
                else
                    plot(FLX/(32*PI), I/app.Ic);
                *phi2p = phi2;
                return (true);
            }
        }
        else {
            phi2 = *phi2p;
        }
        lastz = z;
    }
    return (false);
}


bool
screen::solve32(double phi3, double *phi1p)
{
    double phi1 = *phi1p - PI*1e-4;
    double lastphi = phi1;
    double s3 = vals.iy*sin(phi3);

    double lastz = 0.0;
    for (int count = 0; count < NITER; count++) {

        double phi2;
        if (vals.get2(phi1,phi3,&phi2)) break;
    
        double s1 = vals.ix*sin(phi1);
        double s2 = vals.i2*sin(phi2);

        double f1 = phi1 + PHI*(vals.la*s1 -
            vals.lb*((vals.lfrac-1)*s1 + vals.lfrac*s3)) + app.Order1;

        double f2 = -phi3 + PHI*(vals.lc*(-vals.lfrac*s3 + (1-vals.lfrac)*s1)
            - vals.ld*s3) + app.Order2;

        double f3 = -phi2 - vals.lfrac*PHI*vals.lb*s2;

        double f4 = phi2 + (1-vals.lfrac)*PHI*vals.lc*s2;

        double z = (vals.lc + vals.ld)*f1 - (vals.la + vals.lb)*f2;
        double zz = (vals.lc + vals.ld)*f3 - (vals.la + vals.lb)*f4;

        if (sn(zz) != sn(z)) {
            f1 += f3;
            f2 += f4;
            z += zz;
        }
        else {
            f1 -= f3;
            f2 -= f4;
            z -= zz;
            phi2 = -phi2;
            s2 = -s2;
        }

        if (count > 0) {
            if (z != lastz) {
                double phinew = phi1 - (phi1 - lastphi)*z/(z-lastz);
                LIMIT(phinew);
                lastphi = phi1;
                phi1 = phinew;
            }
            else
                count = NITER;

            if (fabs(z) < .001) {
                if (!vals.solcheck(phi1,phi3)) break;
                double I = s1 + s2 + s3;
                double FLX = f1 + f2;
                if (app.flags.z)
                    plotphi(phi1,phi2,phi3,FLX,I);
                else
                    plot(FLX/(32*PI), I/app.Ic);
                *phi1p = phi1;
                return (true);
            }
        }
        else {
            phi1 += *phi1p;
        }
        lastz = z;
    }
    return (false);
}


bool
screen::solve31(double phi3, double *phi2p)
{
    double phi2 = *phi2p - PI*1e-4;
    double lastphi = phi2;
    double s3 = vals.iy*sin(phi3);

    double lastz = 0.0;
    for (int count = 0; count <= NITER; count++) {

        double phi1;
        if (vals.get1(phi2,phi3,&phi1)) break;
    
        double s1 = vals.ix*sin(phi1);
        double s2 = vals.i2*sin(phi2);

        double f1 = -phi2 - vals.lfrac*PHI*vals.lb*(s2 + s3) + app.Order1;

        double f2 = phi2 - phi3 + PHI*(vals.lc*(-vals.lfrac*s3 +
            (1-vals.lfrac)*s2) - vals.ld*s3) + app.Order2;

        double f3 = phi1 + PHI*(vals.la + (1-vals.lfrac)*vals.lb)*s1;

        double f4 = (1-vals.lfrac)*PHI*vals.lc*s1;

        double z = (vals.lc + vals.ld)*f1 - (vals.la + vals.lb)*f2;
        double zz = (vals.lc + vals.ld)*f3 - (vals.la + vals.lb)*f4;

        if (sn(zz) != sn(z)) {
            f1 += f3;
            f2 += f4;
            z += zz;
        }
        else {
            f1 -= f3;
            f2 -= f4;
            z -= zz;
            phi1 = -phi1;
            s1 = -s1;
        }

        if (count > 0) {
            if (z != lastz) {
                double phinew = phi2 - (phi2 - lastphi)*z/(z-lastz);
                LIMIT(phinew);
                lastphi = phi2;
                phi2 = phinew;
            }
            else
                count = NITER;

            if (fabs(z) < .001) {
                if (!vals.solcheck(phi1,phi3)) break;
                double I = s1 + s2 + s3;
                double FLX = f1 + f2;
                if (app.flags.z)
                    plotphi(phi1,phi2,phi3,FLX,I);
                else
                    plot(FLX/(32*PI), I/app.Ic);
                *phi2p = phi2;
                return (true);
            }
        }
        else {
            phi2 = *phi2p;
        }
        lastz = z;
    }
    return (false);
}


bool
screen::solve21(double phi2, double *phi3p)
{
    double phi3 = *phi3p - PI*1e-4;
    double lastphi = phi3;
    double s2 = vals.i2*sin(phi2);

    double lastz = 0.0;
    for (int count = 0; count <= NITER; count++) {

        double phi1;
        if (vals.get1(phi2,phi3,&phi1)) break;
    
        double s1 = vals.ix*sin(phi1);
        double s3 = vals.iy*sin(phi3);

        double f1 = -phi2 - vals.lfrac*PHI*vals.lb*(s2 + s3) + app.Order1;

        double f2 = phi2 - phi3 + PHI*(vals.lc*(-vals.lfrac*s3 +
            (1-vals.lfrac)*s2) - vals.ld*s3) + app.Order2;

        double f3 = phi1 + PHI*(vals.la + (1-vals.lfrac)*vals.lb)*s1;

        double f4 = (1-vals.lfrac)*PHI*vals.lc*s1;

        double z = (vals.lc + vals.ld)*f1 - (vals.la + vals.lb)*f2;
        double zz = (vals.lc + vals.ld)*f3 - (vals.la + vals.lb)*f4;

        if (sn(zz) != sn(z)) {
            f1 += f3;
            f2 += f4;
            z += zz;
        }
        else {
            f1 -= f3;
            f2 -= f4;
            z -= zz;
            phi1 = -phi1;
            s1 = -s1;
        }

        if (count > 0) {
            if (z != lastz) {
                double phinew = phi3 - (phi3 - lastphi)*z/(z-lastz);
                LIMIT(phinew);
                lastphi = phi3;
                phi3 = phinew;
            }
            else
                count = NITER;

            if (fabs(z) < .001) {
                if (!vals.solcheck(phi1,phi3)) break;
                double I = s1 + s2 + s3;
                double FLX = f1 + f2;
                if (app.flags.z)
                    plotphi(phi1,phi2,phi3,FLX,I);
                else
                    plot(FLX/(32*PI), I/app.Ic);
                *phi3p = phi3;
                return (true);
            }
        }
        else {
            phi3 = *phi3p;
        }
        lastz = z;
    }
    return (false);
}


bool
screen::solve23(double phi2, double *phi1p)
{
    double phi1 = *phi1p - PI*1e-4;
    double lastphi = phi1;
    double s2 = vals.i2*sin(phi2);

    double lastz = 0.0;
    for (int count = 0; count <= NITER; count++) {

        double phi3;
        if (vals.get3(phi1,phi2,&phi3)) break;

        double s1 = vals.ix*sin(phi1);
        double s3 = vals.iy*sin(phi3);

        double f1 = phi1 - phi2 + PHI*(vals.la*s1 -
            vals.lb*((vals.lfrac-1)*s1 + vals.lfrac*s2)) + app.Order1;

        double f2 = phi2 + PHI*((1-vals.lfrac)*vals.lc*(s2 + s1)) + app.Order2;

        double f3 = -vals.lfrac*PHI*vals.lb*s3;

        double f4 = -phi3 -  PHI*(vals.ld + vals.lfrac*vals.lc)*s3;

        double z = (vals.lc + vals.ld)*f1 - (vals.la + vals.lb)*f2;
        double zz = (vals.lc + vals.ld)*f3 - (vals.la + vals.lb)*f4;

        if (sn(zz) != sn(z)) {
            f1 += f3;
            f2 += f4;
            z += zz;
        }
        else {
            f1 -= f3;
            f2 -= f4;
            z -= zz;
            phi3 = -phi3;
            s3 = -s3;
        }

        if (count > 0) {
            if (z != lastz) {
                double phinew = phi1 - (phi1 - lastphi)*z/(z-lastz);
                LIMIT(phinew);
                lastphi = phi1;
                phi1 = phinew;
            }
            else
                count = NITER;

            if (fabs(z) < .001) {
                if (!vals.solcheck(phi1,phi3)) break;
                double I = s1 + s2 + s3;
                double FLX = f1 + f2;
                if (app.flags.z)
                    plotphi(phi1,phi2,phi3,FLX,I);
                else
                    plot(FLX/(32*PI), I/app.Ic);
                *phi1p = phi1;
                return (true);
            }
        }
        else {
            phi1 = *phi1p;
        }
        lastz = z;
    }
    return (false);
}


// Plot threshold points
//
void
screen::plot(double flx, double iy)
{
    int n = - (int) ((flx+1)/.125);
    switch(app.flags.c) {
    case 0:
        set_color(2+abs(n));
        break;
    case 1:
        set_color(13);
        break;
    case 2:
    default:
        set_color(15);
        break;
    }
    flx += n * .125;
    int y = (int) (iy*scale.yy + .5);
    while (flx <= 1) { // clip and plot points
        int x = (int) (flx*scale.xx + .5);
        if (x >= -scale.xmn && x <= scale.xmx && (app.flags.y || y >= 0))
            pixel(scale.xo + x, scale.yo + y);
        if (x <  scale.xmn && x >= -scale.xmx && (app.flags.y || y < 0))
            pixel(scale.xo - x, scale.yo - y);
        flx += .125;
        n++;
        if (!app.flags.c) set_color(2+abs(n));
    }
}


void
screen::plotphi(double phi1, double phi2, double phi3, double flx, double i)
{
    if (app.flags.threej && (app.Order1 != 0 || app.Order2 != 0))
        return;

    if (i == 0 && flx == 0)
        return;
    i = atan2(flx,i);
    int x  = XL + (int)(X*(i + PI)/(2*PI));
    int y1 = YL + (int)(Y*(phi1 + PI)/(2*PI));
    int y2 = YL + (int)(Y*(phi2 + PI)/(2*PI));
    int y3 = YL + (int)(Y*(phi3 + PI)/(2*PI));
    set_color(2);
    pixel(x,y1);
    set_color(3);
    if (app.flags.threej) {
        pixel(x,y2);
        set_color(4);
    }
    pixel(x,y3);

    if (i >= 0)
        i -= PI;
    else
        i += PI;
    phi1 = -phi1;
    phi2 = -phi2;
    phi3 = -phi3;
    x  = XL + (int)(X*(i + PI)/(2*PI));
    y1 = YL + (int)(Y*(phi1 + PI)/(2*PI));
    y2 = YL + (int)(Y*(phi2 + PI)/(2*PI));
    y3 = YL + (int)(Y*(phi3 + PI)/(2*PI));
    set_color(2);
    pixel(x,y1);
    set_color(3);
    if (app.flags.threej) {
        pixel(x,y2);
        set_color(4);
    }
    pixel(x,y3);
}


void
screen::diagram()
{
    clear();
    if (app.flags.threej) {
        int x = app.diag_x - 50;
        int y = app.diag_y - 180;

        set_color(1);

        line(x+100,y+340,x+125,y+340);
        line(x+90,y+350,x+100,y+340);
        line(x+90,y+330,x+100,y+340);

        show_l_icon(x+125,y+317);
        show_x_icon(x+175,y+340);
        show_l_icon(x+225,y+317);
        line(x+275,y+340,x+325,y+340);
        show_l_icon(x+325,y+317);
        show_x_icon(x+375,y+340);
        show_l_icon(x+425,y+317);
        line(x+475,y+340,x+510,y+340);

        line(x+510,y+340,x+500,y+350);
        line(x+510,y+340,x+500,y+330);
        line(x+290,y+400,x+300,y+390);
        line(x+310,y+400,x+300,y+390);

        line(x+300,y+390,x+300,y+370);
        line(x+200,y+300,x+200,y+370);
        line(x+200,y+370,x+400,y+370);
        line(x+400,y+370,x+400,y+300);

        line(x+100,y+300,x+125,y+300);
        show_l_icon(x+125,y+277);
        line(x+175,y+300,x+225,y+300);
        show_l_icon(x+225,y+277);
        line(x+275,y+300,x+325,y+300);
        show_l_icon(x+325,y+277);
        line(x+375,y+300,x+425,y+300);
        show_l_icon(x+425,y+277);
        line(x+475,y+300,x+500,y+300);

        line(x+500,y+300,x+500,y+240);
        line(x+500,y+240,x+100,y+240);
        line(x+100,y+240,x+100,y+300);
        line(x+300,y+240,x+300,y+300);

        line(x+90,y+260,x+110,y+280);
        line(x+90,y+280,x+110,y+260);
        line(x+290,y+260,x+310,y+280);
        line(x+290,y+280,x+310,y+260);
        line(x+490,y+260,x+510,y+280);
        line(x+490,y+280,x+510,y+260);

        line(x+300,y+240,x+300,y+210);
        line(x+280,y+210,x+320,y+210);

        text("Ic",x+100,y+320,0);
        text("I",x+270,y+390,0);
        text("g",x+279,y+387,0);
        text("Ileft",x+180,y+377,0);
        text("Ic1",x+120,y+250,0);
        text("Ic2",x+260,y+250,0);
        text("L1",x+160,y+255,0);
        text("L2",x+220,y+255,0);
        text("L3",x+360,y+255,0);
        text("L4",x+420,y+255,0);
        text("Ic3",x+456,y+250,0);

        set_color(4);
        line(x+60,y+190,x+60,y+420);
        line(x+60,y+420,x+530,y+420);
        line(x+530,y+420,x+530,y+190);
        line(x+530,y+190,x+60,y+190);
    }
    else {
        int x = app.diag_x;
        int y = app.diag_y - 180;

        set_color(1);
        line(x+50,y+350,x+60,y+340);
        line(x+60,y+340,x+175,y+340);
        show_l_icon(x+175,y+317);
        show_x_icon(x+225,y+340);
        show_l_icon(x+275,y+317);
        line(x+325,y+340,x+450,y+340);
        line(x+450,y+340,x+440,y+350);
        line(x+50,y+330,x+60,y+340);
        line(x+450,y+340,x+440,y+330);

        line(x+240,y+400,x+250,y+390);
        line(x+260,y+400,x+250,y+390);
        line(x+250,y+390,x+250,y+300);
        line(x+150,y+300,x+175,y+300);
        show_l_icon(x+175,y+277);
        line(x+225,y+300,x+275,y+300);
        show_l_icon(x+275,y+277);
        line(x+325,y+300,x+350,y+300);
        line(x+350,y+300,x+350,y+240);
        line(x+350,y+240,x+150,y+240);
        line(x+150,y+240,x+150,y+300);

        line(x+140,y+260,x+160,y+280);
        line(x+140,y+280,x+160,y+260);
        line(x+340,y+260,x+360,y+280);
        line(x+340,y+280,x+360,y+260);

        line(x+250,y+240,x+250,y+210);
        line(x+230,y+210,x+270,y+210);

        text("Ic",x+60,y+320,0);
        text("I",x+270,y+390,0);
        text("g",x+279,y+387,0);
        text("Ic1",x+170,y+250,0);
        text("Ic2",x+310,y+250,0);
        text("L1",x+210,y+255,0);
        text("L2",x+270,y+255,0);

        set_color(4);
        line(x+30,y+190,x+30,y+420);
        line(x+30,y+420,x+470,y+420);
        line(x+470,y+420,x+470,y+190);
        line(x+470,y+190,x+30,y+190);
    }
}


char l_icon[24][51] = {
    "00000000000010000000000001000000000000100000000000",
    "00000000000111000000000011100000000001110000000000",
    "00000000001000100000000100010000000010001000000000",
    "00000000001000100000000100010000000010001000000000",
    "00000000010000010000001000001000000100000100000000",
    "00000000010000010000001000001000000100000100000000",
    "00000000100000001000010000000100001000000010000000",
    "00000000100000001000010000000100001000000010000000",
    "00000000100000001000010000000100001000000010000000",
    "00000000100000001000010000000100001000000010000000",
    "00000000100000001000010000000100001000000010000000",
    "00000000100000001000010000000100001000000010000000",
    "00000000100000001000010000000100001000000010000000",
    "00000000100000001000010000000100001000000010000000",
    "00000000100000001000010000000100001000000010000000",
    "00000000010000010000001000001000000100000100000000",
    "00000000010000010000001000001000000100000100000000",
    "00000000001000100000000100010000000010001000000000",
    "00000000001000100000000100010000000010001000000000",
    "00000000000111000000000011100000000001110000000000",
    "00000000000010000000000001000000000000100000000000",
    "00000000001101100000000110110000000011011000000000",
    "00000000110000011000011000001100001100000110000000",
    "11111111000000000111100000000011110000000001111111",
};

char x_icon[7][51] = {
    "11111111111111111111000000000001111111111111111111",
    "00000000000000000001000000000001000000000000000000",
    "00000000000000000001000000000001000000000000000000",
    "00000000000000000001100000000011000000000000000000",
    "00000000000000000000100000000010000000000000000000",
    "00000000000000000000111000001110000000000000000000",
    "00000000000000000000001111111000000000000000000000",
};


// Show inductor symbol
//
void
screen::show_l_icon(int x, int y)
{
    for (int i = 0; i < 24; i++)
        for (int j = 0; j < 50; j++)
            if (l_icon[i][j] - '0') pixel(x+j,y+i);
}


// Show crossover symbol
//
void
screen::show_x_icon(int x, int y)
{
    for (int i = 0; i < 7; i++)
        for (int j = 0; j < 50; j++)
            if (x_icon[i][j] - '0') pixel(x+j,y+i);
}


void
screen::xy_pick()
{
    set_color(0);
    box(XU-38*W,1,XS,YL-1-H3);
    first_ghost = true;
    draw_ghost = true;
    app.flags.rel_x = -1;
    app.flags.rel_y = -1;
}


void
screen::marker(int x, int y)
{
    if (x < XL || x > XU) return;
    if (y < YL || y > YU) return;

    char buf[256];
    if ((app.flags.rel_x < 0) || (app.flags.rel_y < 0)) {
        if (app.flags.z) {
            double x1 = -PI + 2*PI*(x-XL)/X;
            double y1 = -PI + 2*PI*(YS-y-YL)/Y;
            sprintf(buf,"x = %s, y = %s", ecvt12(x1),ecvt12(y1));
        }
        else {
            double x1 = scale.xmin + (scale.xmax - scale.xmin)*(x-XL)/X;
            double y1 = scale.ymin + (scale.ymax - scale.ymin)*(YS-y-YL)/Y;
            sprintf(buf,"Ic = %s, Ig = %s",
                ecvt12(x1/app.Coupling),ecvt12(y1));
        }
    }
    else {
        int px = (int)(X*app.flags.rel_x + XL + .5);
        int py = (int)(Y*app.flags.rel_y + YL + .5);
        if (app.flags.z) {
            double x1 = 2*PI*(x-px)/X;
            double y1 = 2*PI*(y-py)/Y;
            sprintf(buf,"dx = %s, dy = %s", ecvt12(x1),ecvt12(-y1));
        }
        else {
            double x1 = (scale.xmax - scale.xmin)*(x-px)/X;
            double y1 = (scale.ymax - scale.ymin)*(y-py)/Y;
            sprintf(buf,"dIc = %s, dIg = %s",
                ecvt12(x1/app.Coupling),ecvt12(-y1));
        }
    }

#ifdef WIN32_NATIVE
    gc = GetDC(app.scr.da);
    SetROP2(gc, R2_XORPEN);
    y = YS - y;
    set_color(1);
    line(x,YL,x,YU);
    line(XL,y,XU,y);
    SetROP2(gc, R2_COPYPEN);
    set_color(0);
    box(XL,YL-H3-H,XU,YL-H3);
    set_color(1);
    text(buf,XU-strlen(buf)*W,YL-H3-H,0);
#else
    GdkGC *tmpgc = gc;
    gc = xorgc;
    y = YS - y;
    set_color(1);
    line(x,YL,x,YU);
    line(XL,y,XU,y);
    text(buf,XU-strlen(buf)*W,YL-H3-H,0);
    gc = tmpgc;
#endif
}


#define IPMAX(x, y) ((x) > (y) ? (x) : (y))
#define IPMIN(x, y) ((x) < (y) ? (x) : (y))

void
screen::refmark()
{
    if (app.flags.rel_x < 0 || app.flags.rel_y < 0)
        return;
    int x = (int)(X*app.flags.rel_x + XL + .5);
    int y = YS - (int)(Y*app.flags.rel_y + YL + .5);
    int z = W;
#ifdef WIN32_NATIVE
    gc = GetDC(app.scr.da);
    SetROP2(gc, R2_XORPEN);
    set_color(1);
    line(x,IPMAX(YL,y-z),x,IPMIN(YU,y+z));
    line(IPMAX(XL,x-z),y,IPMIN(XU,x+z),y);
    SetROP2(gc, R2_COPYPEN);
#else
    GdkGC *tmpgc = gc;
    gc = xorgc;
    set_color(1);
    line(x,IPMAX(YL,y-z),x,IPMIN(YU,y+z));
    line(IPMAX(XL,x-z),y,IPMIN(XU,x+z),y);
    gc = tmpgc;
#endif
}


// Plot points extracted from files
//
void
screen::plpts()
{
    int i = 5;
    for (extra *p = app.pts; p; p = p->next) {
        set_color(i++);
        for (int j = 0; j < p->size; j++) {
            if (app.flags.f && p->pf[j]) continue;
            if (!app.flags.f && !p->pf[j]) continue;

            int x = (int)( (p->x[j]*app.Coupling -
                scale.xmin)*X/(scale.xmax - scale.xmin) + XL );
            if (x < XL || x > XU)
                return;
            int y = (int)( (p->y[j] - scale.ymin)*Y/(scale.ymax -
                scale.xmin) + YL );
            if (y < YL || y > YU)
                return;
            pixel(x,y+1);
            pixel(x-1,y);
            pixel(x,y);
            pixel(x+1,y);
            pixel(x,y-1);
        }
    }
}
// End of screen functions


bool
values::solcheck(double phi1, double phi3)
{
    double co = PHI*ix*cos(phi1)*(la+lb);
    if (co < -1) return (false);
    co = PHI*iy*cos(phi3)*(lc+ld);
    if (co < -1) return (false);
    return (true);
}


bool
values::get1(double phi2, double phi3, double *phi1)
{
    double z2 = i2*cos(phi2);
    double z3 = iy*cos(phi3);

    double z31 = -PHI*lfrac*lb*z3;
    double z32 = -1 - PHI*(ld + lfrac*lc)*z3;

    double z21 = -1 - PHI*lfrac*lb*z2;
    double z22 = 1 + PHI*(1-lfrac)*lc*z2;

    // z3 + A*z31 + B*z32 = 0
    // z2 + A*z21 + B*z22 = 0
    // B = -(z2 + A*z21)/z22
    // z3 + A*z31 - z32*(z2 + A*z21)/z22 = 0
    // z3 - z32*z2/z22 = A*(-z31 + z32*z21/z22)
    // A = (z3*z22 - z32*z2)/(-z31*z22 + z32*z21)

    double a = (z3*z22 - z32*z2);
    double b = (-z31*z22 + z32*z21);
    double c = -(b*z2 + a*z21);
    double d = b*z22;

    // z1 + a*(1 + PHI*L1*z1 + (1-C)*PHI*L2*z1)/b + c*((1-C)*PHI*L3*z1)/d= 0
    // z1(1 + a*PHI*(L1 + (1-C)*L2)/b + (1-C)*c*PHI*L3/d = -a/b
    // z1 = -a/(b + a*PHI*(L1 + (1-C)*L2) + (1-C)*c*PHI*L3/d)
    // z1 = -a*d/(b*d + a*d*PHI*(L1 + (1-C)*L2) + (1-C)*c*b*PHI*L3)

    double e = b*d + a*d*PHI*(la + (1-lfrac)*lb) + c*b*PHI*(1-lfrac)*lc;
    double z1 = -a*d;

    if (e == 0) {
        if (z1 == 0) {
            *phi1 = .5*PI;
            return (false);
        }
        return (true);
    }

    z1 /= e*ix;

    if (fabs(z1) <= 1) {
        *phi1 = acos(z1);
        return (false);
    }
    return (true);
}


bool
values::get2(double phi1, double phi3, double *phi2)
{
    double z1 = ix*cos(phi1);
    double z3 = iy*cos(phi3);

    double z11 = 1 + PHI*(la + (1-lfrac)*lb)*z1;
    double z12 = PHI*(1-lfrac)*lc*z1;

    double z31 = -PHI*lfrac*lb*z3;
    double z32 = -1 - PHI*(ld + lfrac*lc)*z3;

    // z1 + A*z11 + B*z12 = 0
    // z3 + A*z31 + B*z32 = 0
    // B = -(z3 + A*z31)/z32
    // z1 + A*z11 - z12*(z3 + A*z31)/z32 = 0
    // z1 - z12*z3/z32 = A*(-z11 + z12*z31/z32)
    // A = (z1*z32 - z12*z3)/(-z11*z32 + z12*z31);

    double a = (z1*z32 - z12*z3);
    double b = (-z11*z32 + z12*z31);
    double c = -(b*z3 + a*z31);
    double d = b*z32;

    // z2 - a*(1 + C*PHI*L2*z2)/b + c*(1 + (1-C)*PHI*L3*z2)/d = 0
    // z2(1 - C*a*PHI*L2/b + (1-C)*c*PHI*L3/d) = a/b - c/d
    // z2 = (a*d - c*b)/(b*d - C*a*d*PHI*L2 + (1-C)*c*b*PHI*L3)

    double e = b*d - a*d*PHI*lfrac*lb + c*b*PHI*(1-lfrac)*lc;
    double z2 = a*d - c*b;

    if (e == 0) {
        if (z3 == 0) {
            *phi2 = .5*PI;
            return (false);
        }
        return (true);
    }

    z2 /= e*i2;

    if (fabs(z2) <= 1) {
        *phi2 = acos(z2);
        return (false);
    }
    return (true);
}


bool
values::get3(double phi1,double phi2, double *phi3)
{
    double z1 = ix*cos(phi1);
    double z2 = i2*cos(phi2);

    double z11 = 1 + PHI*(la + (1-lfrac)*lb)*z1;
    double z12 = PHI*(1-lfrac)*lc*z1;

    double z21 = -1 - PHI*lfrac*lb*z2;
    double z22 = 1 + PHI*(1-lfrac)*lc*z2;

    // z1 + A*z11 + B*z12 = 0
    // z2 + A*z21 + B*z22 = 0
    // B = -(z2 + A*z21)/z22
    // z1 + A*z11 - z12*(z2 + A*z21)/z22 = 0
    // z1 - z12*z2/z22 = A*(-z11 + z12*z21/z22)
    // A = (z1*z22 - z12*z2)/(-z11*z22 + z12*z21);

    double a = (z1*z22 - z12*z2);
    double b = (-z11*z22 + z12*z21);
    double c = -(b*z2 + a*z21);
    double d = b*z22;

    // z3 - a*(C*PHI*L2*z3)/b - c*(1 + PHI*L4*z3 + C*PHI*L3*z3)/d = 0
    // z3(1 - C*a*PHI*L2/b - c*PHI*(L4 + C*L3)/d) = c/d
    // z3 = c/(d - C*a*PHI*L2*d/b - c*PHI*(L4 + C*L3))
    // z3 = c*b/(b*d - C*a*PHI*L2*d - c*b*PHI*(L4 + C*L3))

    double e = b*d - a*d*PHI*lfrac*lb - c*b*PHI*(ld + lfrac*lc);
    double z3 = c*b;

    if (e == 0) {
        if (z3 == 0) {
            *phi3 = .5*PI;
            return (false);
        }
        return (true);
    }

    z3 /= e*iy;

    if (fabs(z3) <= 1) {
        *phi3 = acos(z3);
        return (false);
    }
    return (true);
}


// Format: space or +, x.xxxxx, e, + or -, xx
// keeps track of 10 calls for use in printf.
//
static char *
ecvt12(double d)
{
    static char s[10][16];
    static int num;

    num++;
    if (num > 9) num = 0;
    sprintf(s[num],"%+.5e",d);
    char *ss = strchr(s[num],'e');
    ss++;
    int xp = atoi(ss);

    if (xp > 99) xp = 99;
    if (xp < -99) {
        sprintf(s[num]," 0.00000e+00");
        return s[num];
    }

    *(s[num]+8) = 'e';
    if (xp < 0)    {
        *(s[num]+9) = '-';
        xp = -xp;
    }
    else
        *(s[num]+9) = '+';
    *(s[num]+10) = '0' + (xp/10);
    *(s[num]+11) = '0' + (xp%10);
    *(s[num]+12) = '\0';
    if (*s[num] == '+')
        *s[num] = ' ';
    return (s[num]);
}


